"""Ambiguous Groups anti-pattern detector class"""
import re

from sqleyes.detector.antipatterns.abstract_base_class import AbstractDetector
from sqleyes.definitions.definitions import DEFINITIONS
from sqleyes.detector.detector_output import DetectorOutput
from sqleyes.utils.query_functions import (check_single_value_rule,
                                           get_columns_from_group_by_statement,
                                           get_columns_from_select_statement)


class AmbiguousGroupsDetector(AbstractDetector):

    filename = DEFINITIONS["anti_patterns"]["ambiguous_groups"]["filename"]
    type = DEFINITIONS["anti_patterns"]["ambiguous_groups"]["type"]
    title = DEFINITIONS["anti_patterns"]["ambiguous_groups"]["title"]

    def __init__(self, query):
        super().__init__(query)

    def check(self):
        pattern = re.compile(r'GROUP\s*BY', re.IGNORECASE)

        if pattern.search(self.query):
            # GROUP BY pattern is found in the query

            # Get columns in SELECT & GROUP BY statement
            select_columns = get_columns_from_select_statement(self.query)
            group_columns = get_columns_from_group_by_statement(self.query)

            # Get columns which are in select_columns but not in group_columns
            remaining_columns = list(set(select_columns) - set(group_columns))

            # Check if the remaining columns break single-value rule
            single_values = check_single_value_rule(remaining_columns)

            if not single_values:
                return DetectorOutput(certainty="high",
                                      description=super().get_description(),
                                      detector_type=self.detector_type,
                                      title=self.title,
                                      type=self.type)

            return None

        return None



"""Detector class running various detectors"""
from typing import List
from sqleyes.detector.antipatterns.ambiguous_groups import AmbiguousGroupsDetector
from sqleyes.detector.antipatterns.fear_of_the_unknown import FearOfTheUnknownDetector
from sqleyes.detector.antipatterns.implicit_columns import ImplicitColumnsDetector
from sqleyes.detector.antipatterns.poor_mans_search_engine import PoorMansSearchEngineDetector
from sqleyes.detector.antipatterns.random_selection import RandomSelectionDetector
from sqleyes.detector.detector_output import DetectorOutput


class Detector:
    """
    This is a Detector class that is responsible for detecting errors
    in a given query.

    Attributes:
        query (str): The query to be analyzed.
    """

    def __init__(self, query: str):
        self.query = query
        self.anti_pattern_list: List[DetectorOutput] = []

    def run(self) -> List[DetectorOutput]:
        """
        This function runs various detectors on a query.

        Returns:
            List[DetectorOutput]: A list of Detector outputs of various
            detectors.
        """
        ap_ambiguous_groups = AmbiguousGroupsDetector(query=self.query) \
            .check()
        self.anti_pattern_list.append(ap_ambiguous_groups)

        ap_fear_of_the_unknown = FearOfTheUnknownDetector(query=self.query) \
            .check()
        self.anti_pattern_list.append(ap_fear_of_the_unknown)

        ap_implicit_col = ImplicitColumnsDetector(query=self.query).check()
        self.anti_pattern_list.append(ap_implicit_col)

        ap_pm_search_engine = PoorMansSearchEngineDetector(query=self.query) \
            .check()
        self.anti_pattern_list.append(ap_pm_search_engine)

        ap_random_selection = RandomSelectionDetector(query=self.query).check()
        self.anti_pattern_list.append(ap_random_selection)

        return [ap for ap in self.anti_pattern_list if ap is not None]



# Setup argument parser
import argparse

from sqleyes.main import main
from sqleyes.printer.printer import IntroPrinter, OutputPrinter


parser = argparse.ArgumentParser(
    description="Analyze raw SQL queries for anti-patterns")

parser.add_argument('-q', '--query', metavar="", type=str, required=True,
                    help="A raw SQL query to analyze")

parser.add_argument('-d', '--description', action="store_true",
                    help="Show descriptions of found errors")

parser.set_defaults(description=False)

args = parser.parse_args()


def cli():
    IntroPrinter(args.query).print()
    output = main(args.query)
    OutputPrinter(output).print(args.description)


if __name__ == '__main__':
    IntroPrinter(args.query).print()
    output = main(args.query)
    OutputPrinter(output).print(args.description)



"""Abstract anti-pattern detector class"""
from abc import ABC, abstractmethod

from sqleyes.utils.load_file import load_description


class AbstractDetector(ABC):
    """
    This is a class for detecting anti-patterns.

    Parameters:
        query : str
            The query to be searched for.

    Attributes:
        detector_type : str
            The type of detector.
        query : str
            The query to be searched for.
    """
    filename: str = NotImplemented
    type: str = NotImplemented
    title: str = NotImplemented

    @abstractmethod
    def __init__(self, query):
        self.detector_type = "anti-pattern"
        self.query = query

    @abstractmethod
    def check(self):
        pass

    def get_description(self):
        return load_description("sqleyes.definitions", "antipatterns/",
                                self.filename)



"""Detector Ouput class specifying general output format"""
import json


class DetectorOutput(object):
    """
    This class represents the output of a detector.

    Attributes:
        certainty (str): The certainty of the detector detecting the AP.
        detector_type (str): The type of detector that produced this output.
        title (str): The title of output.
        type (str): The type of output.
    """
    def __init__(self, certainty: str, description: str, detector_type: str,
                 title: str, type: str):
        if certainty not in ["low", "medium", "high"]:
            raise Exception("Certainty must be specified as either 'low', \
                             'medium' or 'high' ")
        self.certainty = certainty
        self.description = description
        self.detector_type = detector_type
        self.title = title
        self.type = type
        self.dict = self.__create_dictionary()

    def __create_dictionary(self):
        return {
            "certainty": self.certainty,
            "description": self.description,
            "detector_type": self.detector_type,
            "title": self.title,
            "type": self.type
        }

    def __getitem__(self, item):
        return self.dict[item]

    def __repr__(self):
        return json.dumps(self.__create_dictionary())

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return (self.type == other.type
                    and self.detector_type == other.detector_type)



"""Poor Man's Search Engine anti-pattern detector class"""
import re
from sqleyes.detector.antipatterns.abstract_base_class import AbstractDetector
from sqleyes.definitions.definitions import DEFINITIONS
from sqleyes.detector.detector_output import DetectorOutput


class PoorMansSearchEngineDetector(AbstractDetector):

    filename = DEFINITIONS["anti_patterns"]["poor_mans_search_engine"]["filename"]
    type = DEFINITIONS["anti_patterns"]["poor_mans_search_engine"]["type"]
    title = DEFINITIONS["anti_patterns"]["poor_mans_search_engine"]["title"]

    def __init__(self, query):
        super().__init__(query)

    def check(self):
        patterns = [re.compile("(LIKE)", re.IGNORECASE),
                    re.compile("(REGEXP)", re.IGNORECASE)]

        for pattern in patterns:
            if pattern.search(self.query):
                return DetectorOutput(certainty="medium",
                                      description=super().get_description(),
                                      detector_type=self.detector_type,
                                      title=self.title,
                                      type=self.type)

        return None



"""A Python CLI tool for detecting anti-patterns in raw SQL queries"""

from sqleyes.detector.detector import Detector


def main(query: str):
    """
    This function bootstraps the detector and runs it

    Parameters:
        query (str): A raw SQL query.

    Returns:
        dict: A dictionary of detected anti-patterns.
    """
    return Detector(query).run()



from abc import ABC, abstractmethod
from typing import List
import pkg_resources
import sqlparse
from rich.console import Console
from rich.markdown import Markdown
from rich.padding import Padding
from rich.table import Table
from sqleyes.detector.detector_output import DetectorOutput


class AbstractPrinter(ABC):
    """
    This is a class for printing to console.

    Parameters:
        none

    Attributes:
        console : object
            The rich console object
    """

    @abstractmethod
    def __init__(self):
        self.console = Console()

    @abstractmethod
    def print(self):
        pass


class IntroPrinter(AbstractPrinter):
    def __init__(self, input_query):
        super().__init__()
        self.input_query = sqlparse.format(input_query, reindent=True,
                                           keyword_case='upper')

    def print(self):
        self.console.print(f"[bold cyan]SQLEyes v{pkg_resources.require('sqleyes')[0].version}[/bold cyan]")
        self.console.print()


class OutputPrinter(AbstractPrinter):
    def __init__(self, detector_output: List[DetectorOutput]):
        super().__init__()
        self.detector_output = detector_output

    def print_summary(self):
        table = Table(title=f"[bold cyan]Summary of analysis[/bold cyan] \nFound {len(self.detector_output)} errors in the given query", title_justify="left")

        table.add_column("Error", justify="right", style="cyan", no_wrap=True)
        table.add_column("Title", style="red")
        table.add_column("Type", style="red")
        table.add_column("Certainty", justify="right", style="green")

        for output in self.detector_output:
            table.add_row(output["detector_type"], output["title"], output["type"], output["certainty"])

        self.console.print(table)

    def print_descriptions(self):
        self.console.print("[bold cyan]Detailed descriptions of found errors[/bold cyan]")

        for output in self.detector_output:
            type = output["type"]
            title = output["title"]
            filename = output["type"].replace(" ", "_").replace("'", "").lower()

            self.console.print("-"*10, style="underline")

            with open(pkg_resources.resource_filename("sqleyes.definitions", f"antipatterns/{filename}.md"), "r+") as definition:
                self.console.print(f"[bold red]type[/bold red]: {type}")
                self.console.print(f"[bold red]title[/bold red]: {title}")
                self.console.print("[bold red]Description[/bold red]:")
                self.console.print(Padding(Markdown(definition.read()), (1, 2)))
                self.console.print("-"*10, style="underline")

    def print(self, descriptions=False):
        self.print_summary()
        if descriptions:
            self.console.print()
            self.print_descriptions()



"""Implicit Columns anti-pattern detector class"""
import re
from sqleyes.detector.antipatterns.abstract_base_class import AbstractDetector
from sqleyes.definitions.definitions import DEFINITIONS
from sqleyes.detector.detector_output import DetectorOutput


class ImplicitColumnsDetector(AbstractDetector):

    filename = DEFINITIONS["anti_patterns"]["implicit_columns"]["filename"]
    type = DEFINITIONS["anti_patterns"]["implicit_columns"]["type"]
    title = DEFINITIONS["anti_patterns"]["implicit_columns"]["title"]

    def __init__(self, query):
        super().__init__(query)

    def check(self):
        pattern = re.compile("(SELECT\\s+\\*)", re.IGNORECASE)

        if pattern.search(self.query):
            return DetectorOutput(certainty="high",
                                  description=super().get_description(),
                                  detector_type=self.detector_type,
                                  title=self.title,
                                  type=self.type)

        return None



DEFINITIONS = {
    "anti_patterns": {
        "ambiguous_groups": {
            "filename": "ambiguous_groups.md",
            "title": "Incorrect GROUP BY usage",
            "type": "Ambiguous Groups"
        },
        "fear_of_the_unknown": {
            "filename": "fear_of_the_unknown.md",
            "title": "Incorrect NULL usage",
            "type": "Fear of the Unknown"
        },
        "implicit_columns": {
            "filename": "implicit_columns.md",
            "title": "Avoid usage of wildcard selector",
            "type": "Implicit Columns"
        },
        "poor_mans_search_engine": {
            "filename": "poor_mans_search_engine.md",
            "title": "Avoid pattern matching",
            "type": "Poor Man's Search Engine"
        },
        "random_selection": {
            "filename": "random_selection.md",
            "title": "Avoid ORDER BY RAND() usage",
            "type": "Random Selection"
        }
    }
}



"""Fear of the Unknown anti-pattern detector class"""
import re
from sqleyes.detector.antipatterns.abstract_base_class import AbstractDetector
from sqleyes.definitions.definitions import DEFINITIONS
from sqleyes.detector.detector_output import DetectorOutput


class FearOfTheUnknownDetector(AbstractDetector):

    filename = DEFINITIONS["anti_patterns"]["fear_of_the_unknown"]["filename"]
    type = DEFINITIONS["anti_patterns"]["fear_of_the_unknown"]["type"]
    title = DEFINITIONS["anti_patterns"]["fear_of_the_unknown"]["title"]

    def __init__(self, query):
        super().__init__(query)

    def check(self):
        patterns = [re.compile(r'<>\s*NULL', re.IGNORECASE),
                    re.compile(r'!=\s*NULL', re.IGNORECASE),
                    re.compile(r'=\s*NULL', re.IGNORECASE)]
        for pattern in patterns:
            if pattern.search(self.query):
                return DetectorOutput(certainty="high",
                                      description=super().get_description(),
                                      detector_type=self.detector_type,
                                      title=self.title,
                                      type=self.type)

        return None



"""Utility functions w.r.t queries"""
import re
from typing import List

import sqlparse

from sqleyes.utils.query_keywords import SQL_FUNCTIONS


def get_columns_from_select_statement(query: str) -> List[str]:
    """
    This function takes a query string as input and returns a list of columns
    in the SELECT statement.

    Parameters:
        query (str): The query string.

    Returns:
        columns (list): A list of columns selected in the SELECT statement.
    """
    columns = re.findall(r'SELECT (.*?) FROM', query,
                         flags=re.DOTALL | re.IGNORECASE)

    if len(columns) == 0:
        return []

    columns = columns[0].split(',')
    columns = [column.strip() for column in columns]
    return columns


def get_columns_from_group_by_statement(query: str) -> List[str]:
    """
    This function takes a query string as input and returns a list of columns
    in the GROUP BY statement.

    Parameters:
        query (str): A SQL query string.

    Returns:
        List[str]: A list of column names in the GROUP BY statement.
    """
    tokens = sqlparse.parse(query)[0].tokens

    # Find index of group by keyword in tokens
    for i in range(0, len(tokens)):
        if tokens[i].value.upper() == "GROUP BY":
            break

    # Query has no GROUP BY statement
    if i == len(tokens) - 1:
        return []

    # Find possible index of next keyword
    for j in range(i + 1, len(tokens)):
        if tokens[j].ttype is sqlparse.tokens.Keyword:
            break

    # Get column names
    group_columns = []
    for item in tokens[i:j + 1]:
        if isinstance(item, sqlparse.sql.IdentifierList):
            for identifier in item.get_identifiers():
                group_columns.append(identifier.get_name())
        elif isinstance(item, sqlparse.sql.Identifier):
            group_columns.append(item.get_name())

    return group_columns


def check_single_value_rule(columns: List[str]) -> bool:
    """
    This function checks if the columns in the list break the single-value
    rule in SQL.

    Parameters:
        columns (List[str]): A list of columns.

    Returns:
        bool: True if all columns conform to the single value rule,
        False otherwise.
    """
    for column in columns:
        single_value = False

        if column == "*":
            single_value = True

        for function in SQL_FUNCTIONS:
            if re.search(function, column, re.IGNORECASE):
                single_value = True

        if not single_value:
            return False

    return True



import pkg_resources


def load_description(package: str, path: str, filename: str):
    """
    This function loads a static description file.

    Parameters:
        package (str): Package name where file is located in
        path (str): Path within the package.
        filename (str): Name of file to load.

    Returns:
        str: Content of loaded file
    """
    with open(pkg_resources.resource_filename(
            package, f"{path}{filename}"), "r+") as file:
        return file.read()
